# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from sparkmagic.livyclientlib.sendtosparkcommand import SendToSparkCommand
from sparkmagic.livyclientlib.command import Command
from sparkmagic.livyclientlib.exceptions import BadUserDataException


class SendStringToSparkCommand(SendToSparkCommand):
    def _scala_command(
        self, input_variable_name, input_variable_value, output_variable_name
    ):
        self._assert_input_is_string_type(input_variable_name, input_variable_value)
        scala_code = 'var {} = """{}"""'.format(
            output_variable_name, input_variable_value
        )
        return Command(scala_code)

    def _pyspark_command(
        self, input_variable_name, input_variable_value, output_variable_name
    ):
        self._assert_input_is_string_type(input_variable_name, input_variable_value)
        pyspark_code = "{} = {}".format(
            output_variable_name, repr(input_variable_value)
        )
        return Command(pyspark_code)

    def _r_command(
        self, input_variable_name, input_variable_value, output_variable_name
    ):
        self._assert_input_is_string_type(input_variable_name, input_variable_value)
        escaped_input_variable_value = input_variable_value.replace(
            "\\", "\\\\"
        ).replace('"', '\\"')
        r_code = """assign("{}","{}")""".format(
            output_variable_name, escaped_input_variable_value
        )
        return Command(r_code)

    def _assert_input_is_string_type(self, input_variable_name, input_variable_value):
        if not isinstance(input_variable_value, str):
            wrong_type = input_variable_value.__class__.__name__
            raise BadUserDataException(
                "{} is not a str or bytes! Got {} instead".format(
                    input_variable_name, wrong_type
                )
            )



# Copyright (c) 2015  aggftw@gmail.com
# Distributed under the terms of the Modified BSD License.

import os

HOME_PATH = os.environ.get("SPARKMAGIC_CONF_DIR", "~/.sparkmagic")
CONFIG_FILE = os.environ.get("SPARKMAGIC_CONF_FILE", "config.json")

SESSION_KIND_SPARK = "spark"
SESSION_KIND_PYSPARK = "pyspark"
SESSION_KIND_SPARKR = "sparkr"
SESSION_KINDS_SUPPORTED = [
    SESSION_KIND_SPARK,
    SESSION_KIND_PYSPARK,
    SESSION_KIND_SPARKR,
]

LIBRARY_LOADED_EVENT = "notebookLoaded"
CLUSTER_CHANGE_EVENT = "notebookClusterChange"
SESSION_CREATION_START_EVENT = "notebookSessionCreationStart"
SESSION_CREATION_END_EVENT = "notebookSessionCreationEnd"
SESSION_DELETION_START_EVENT = "notebookSessionDeletionStart"
SESSION_DELETION_END_EVENT = "notebookSessionDeletionEnd"
STATEMENT_EXECUTION_START_EVENT = "notebookStatementExecutionStart"
STATEMENT_EXECUTION_END_EVENT = "notebookStatementExecutionEnd"
SQL_EXECUTION_START_EVENT = "notebookSqlExecutionStart"
SQL_EXECUTION_END_EVENT = "notebookSqlExecutionEnd"
MAGIC_EXECUTION_START_EVENT = "notebookMagicExecutionStart"
MAGIC_EXECUTION_END_EVENT = "notebookMagicExecutionEnd"

CLUSTER_DNS_NAME = "ClusterDnsName"
SESSION_ID = "SessionId"
SESSION_GUID = "SessionGuid"
STATEMENT_ID = "StatementId"
STATEMENT_GUID = "StatementGuid"
SQL_GUID = "SqlGuid"
MAGIC_NAME = "MagicName"
MAGIC_GUID = "MagicGuid"
LIVY_KIND = "LivyKind"
STATUS = "Status"
SUCCESS = "Success"
EXCEPTION_TYPE = "ExceptionType"
EXCEPTION_MESSAGE = "ExceptionMessage"
SAMPLE_METHOD = "SampleMethod"
MAX_ROWS = "MaxRows"
SAMPLE_FRACTION = "SampleFraction"
ERROR_MESSAGE = "ErrorMessage"
STATUS_CODE = "StatusCode"

CONTEXT_NAME_SPARK = "spark"
CONTEXT_NAME_SQL = "sql"

LANG_SCALA = "scala"
LANG_PYTHON = "python"
LANG_R = "r"
LANGS_SUPPORTED = [LANG_SCALA, LANG_PYTHON, LANG_R]

LONG_RANDOM_VARIABLE_NAME = "yQeKOYBsFgLWWGWZJu3y"

WIDGET_WIDTH = "800px"

MAGICS_LOGGER_NAME = "magicsLogger"

# The list here https://livy.incubator.apache.org/docs/latest/rest-api.html
# appears incomplete; full list is in
# https://github.com/apache/incubator-livy/blob/master/core/src/main/scala/org/apache/livy/sessions/SessionState.scala:
IDLE_SESSION_STATUS = "idle"
ERROR_SESSION_STATUS = "error"
DEAD_SESSION_STATUS = "dead"
NOT_STARTED_SESSION_STATUS = "not_started"
STARTING_SESSION_STATUS = "starting"
BUSY_SESSION_STATUS = "busy"
SUCCESS_SESSION_STATUS = "success"
SHUT_DOWN_SESSION_STATUS = "shutting_down"
RUNNING_SESSION_STATUS = "running"
KILLED_SESSION_STATUS = "killed"
RECOVERING_SESSION_STATUS = "recovering"

POSSIBLE_SESSION_STATUS = [
    NOT_STARTED_SESSION_STATUS,
    IDLE_SESSION_STATUS,
    STARTING_SESSION_STATUS,
    BUSY_SESSION_STATUS,
    ERROR_SESSION_STATUS,
    DEAD_SESSION_STATUS,
    SUCCESS_SESSION_STATUS,
    SHUT_DOWN_SESSION_STATUS,
    RUNNING_SESSION_STATUS,
    KILLED_SESSION_STATUS,
    RECOVERING_SESSION_STATUS,
]
FINAL_STATUS = [
    DEAD_SESSION_STATUS,
    ERROR_SESSION_STATUS,
    SUCCESS_SESSION_STATUS,
    KILLED_SESSION_STATUS,
]

ERROR_STATEMENT_STATUS = "error"
CANCELLED_STATEMENT_STATUS = "cancelled"
AVAILABLE_STATEMENT_STATUS = "available"
FINAL_STATEMENT_STATUS = [
    ERROR_STATEMENT_STATUS,
    CANCELLED_STATEMENT_STATUS,
    AVAILABLE_STATEMENT_STATUS,
]

DELETE_SESSION_ACTION = "delete"
START_SESSION_ACTION = "start"
DO_NOTHING_ACTION = "nothing"

INTERNAL_ERROR_MSG = (
    "An internal error was encountered.\n"
    "Please file an issue at https://github.com/jupyter-incubator/sparkmagic\nError:\n{}"
)
EXPECTED_ERROR_MSG = "An error was encountered:\n{}"

YARN_RESOURCE_LIMIT_MSG = "Queue's AM resource limit exceeded."
RESOURCE_LIMIT_WARNING = (
    "Warning: The Spark session does not have enough YARN resources to start. {}"
)
COMMAND_INTERRUPTED_MSG = "Interrupted by user"
COMMAND_CANCELLATION_FAILED_MSG = (
    "Interrupted by user but Livy failed to cancel the Spark statement. "
    "The Livy session might have become unusable."
)

LIVY_HEARTBEAT_TIMEOUT_PARAM = "heartbeatTimeoutInSecond"
LIVY_KIND_PARAM = "kind"

NO_AUTH = "None"
AUTH_KERBEROS = "Kerberos"
AUTH_BASIC = "Basic_Access"

CONFIGURABLE_RETRY = "configurable"
LINEAR_RETRY = "linear"

MIMETYPE_IMAGE_PNG = "image/png"
MIMETYPE_TEXT_HTML = "text/html"
MIMETYPE_TEXT_PLAIN = "text/plain"



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from sparkmagic.livyclientlib.command import Command
from sparkmagic.livyclientlib.exceptions import BadUserDataException
import sparkmagic.utils.constants as constants

from abc import abstractmethod


class SendToSparkCommand(Command):
    def __init__(
        self,
        input_variable_name,
        input_variable_value,
        output_variable_name,
        spark_events=None,
    ):
        super(SendToSparkCommand, self).__init__("", spark_events)
        self.input_variable_name = input_variable_name
        self.input_variable_value = input_variable_value
        self.output_variable_name = output_variable_name

    def execute(self, session):
        try:
            command = self.to_command(
                session.kind,
                self.input_variable_name,
                self.input_variable_value,
                self.output_variable_name,
            )
            return command.execute(session)
        except Exception as e:
            raise e

    def to_command(
        self, kind, input_variable_name, input_variable_value, output_variable_name
    ):
        if kind == constants.SESSION_KIND_PYSPARK:
            return self._pyspark_command(
                input_variable_name, input_variable_value, output_variable_name
            )
        elif kind == constants.SESSION_KIND_SPARK:
            return self._scala_command(
                input_variable_name, input_variable_value, output_variable_name
            )
        elif kind == constants.SESSION_KIND_SPARKR:
            return self._r_command(
                input_variable_name, input_variable_value, output_variable_name
            )
        else:
            raise BadUserDataException("Kind '{}' is not supported.".format(kind))

    @abstractmethod
    def _scala_command(
        self, input_variable_name, input_variable_value, output_variable_name
    ):
        raise NotImplementedError  # override and provide proper implementation in supertype!

    @abstractmethod
    def _pyspark_command(
        self, input_variable_name, input_variable_value, output_variable_name
    ):
        raise NotImplementedError  # override and provide proper implementation in supertype!

    @abstractmethod
    def _r_command(
        self, input_variable_name, input_variable_value, output_variable_name
    ):
        raise NotImplementedError  # override and provide proper implementation in supertype!



from sparkmagic.utils.constants import LANG_PYTHON, LANG_SCALA, LANG_R
from sparkmagic.kernels.sparkkernel.sparkkernel import SparkKernel
from sparkmagic.kernels.pysparkkernel.pysparkkernel import PySparkKernel
from sparkmagic.kernels.sparkrkernel.sparkrkernel import SparkRKernel


class TestPyparkKernel(PySparkKernel):
    def __init__(self):
        kwargs = {"testing": True}
        super(TestPyparkKernel, self).__init__(**kwargs)


class TestSparkKernel(SparkKernel):
    def __init__(self):
        kwargs = {"testing": True}
        super(TestSparkKernel, self).__init__(**kwargs)


class TestSparkRKernel(SparkRKernel):
    def __init__(self):
        kwargs = {"testing": True}
        super(TestSparkRKernel, self).__init__(**kwargs)


def test_pyspark_kernel_configs():
    kernel = TestPyparkKernel()
    assert kernel.session_language == LANG_PYTHON

    assert kernel.implementation == "PySpark"
    assert kernel.language == LANG_PYTHON
    assert kernel.language_version == "0.1"
    assert kernel.language_info == {
        "name": "pyspark",
        "mimetype": "text/x-python",
        "codemirror_mode": {"name": "python", "version": 3},
        "file_extension": ".py",
        "pygments_lexer": "python3",
    }


def test_spark_kernel_configs():
    kernel = TestSparkKernel()

    assert kernel.session_language == LANG_SCALA

    assert kernel.implementation == "Spark"
    assert kernel.language == LANG_SCALA
    assert kernel.language_version == "0.1"
    assert kernel.language_info == {
        "name": "scala",
        "mimetype": "text/x-scala",
        "pygments_lexer": "scala",
        "file_extension": ".sc",
        "codemirror_mode": "text/x-scala",
    }


def test_sparkr_kernel_configs():
    kernel = TestSparkRKernel()

    assert kernel.session_language == LANG_R

    assert kernel.implementation == "SparkR"
    assert kernel.language == LANG_R
    assert kernel.language_version == "0.1"
    assert kernel.language_info == {
        "name": "sparkR",
        "mimetype": "text/x-rsrc",
        "pygments_lexer": "r",
        "file_extension": ".r",
        "codemirror_mode": "text/x-rsrc",
    }



import atexit
from mock import MagicMock, PropertyMock
from nose.tools import raises, assert_equals

import sparkmagic.utils.configuration as conf
from sparkmagic.livyclientlib.exceptions import SessionManagementException
from sparkmagic.livyclientlib.sessionmanager import SessionManager


@raises(SessionManagementException)
def test_get_client_throws_when_client_not_exists():
    manager = get_session_manager()

    manager.get_session("name")


def test_get_client():
    client = MagicMock()
    manager = get_session_manager()

    manager.add_session("name", client)

    assert_equals(client, manager.get_session("name"))


@raises(SessionManagementException)
def test_delete_client():
    client = MagicMock()
    manager = get_session_manager()

    manager.add_session("name", client)
    manager.delete_client("name")

    manager.get_session("name")


@raises(SessionManagementException)
def test_delete_client_throws_when_client_not_exists():
    manager = get_session_manager()

    manager.delete_client("name")


@raises(SessionManagementException)
def test_add_client_throws_when_client_exists():
    client = MagicMock()
    manager = get_session_manager()

    manager.add_session("name", client)
    manager.add_session("name", client)


def test_client_names_returned():
    client = MagicMock()
    manager = get_session_manager()

    manager.add_session("name0", client)
    manager.add_session("name1", client)

    assert_equals({"name0", "name1"}, set(manager.get_sessions_list()))


def test_get_any_client():
    client = MagicMock()
    manager = get_session_manager()

    manager.add_session("name", client)

    assert_equals(client, manager.get_any_session())


@raises(SessionManagementException)
def test_get_any_client_raises_exception_with_no_client():
    manager = get_session_manager()

    manager.get_any_session()


@raises(SessionManagementException)
def test_get_any_client_raises_exception_with_two_clients():
    client = MagicMock()
    manager = get_session_manager()
    manager.add_session("name0", client)
    manager.add_session("name1", client)

    manager.get_any_session()


def test_clean_up():
    client0 = MagicMock()
    client1 = MagicMock()
    manager = get_session_manager()
    manager.add_session("name0", client0)
    manager.add_session("name1", client1)

    manager.clean_up_all()

    client0.delete.assert_called_once_with()
    client1.delete.assert_called_once_with()


def test_cleanup_all_sessions_on_exit():
    conf.override(conf.cleanup_all_sessions_on_exit.__name__, True)
    client0 = MagicMock()
    client1 = MagicMock()
    manager = get_session_manager()
    manager.add_session("name0", client0)
    manager.add_session("name1", client1)

    atexit._run_exitfuncs()

    client0.delete.assert_called_once_with()
    client1.delete.assert_called_once_with()
    manager.ipython_display.writeln.assert_called_once_with(
        "Cleaning up livy sessions on exit is enabled"
    )


def test_cleanup_all_sessions_on_exit_fails():
    """
    Cleanup on exit is best effort only. When cleanup fails, exception is caught and error is logged.
    """
    conf.override(conf.cleanup_all_sessions_on_exit.__name__, True)
    client0 = MagicMock()
    client1 = MagicMock()
    client0.delete.side_effect = Exception("Mocked exception for client1.delete")
    manager = get_session_manager()
    manager.add_session("name0", client0)
    manager.add_session("name1", client1)

    atexit._run_exitfuncs()

    client0.delete.assert_called_once_with()
    client1.delete.assert_not_called()


def test_get_session_id_for_client():
    manager = get_session_manager()
    manager.get_sessions_list = MagicMock(return_value=["name"])
    manager._sessions["name"] = MagicMock()

    id = manager.get_session_id_for_client("name")

    assert id is not None


def test_get_session_name_by_id_endpoint():
    manager = get_session_manager()
    id_to_search = "0"
    endpoint_to_search = "endpoint"
    name_to_search = "name"

    name = manager.get_session_name_by_id_endpoint(id_to_search, endpoint_to_search)
    assert_equals(None, name)

    session = MagicMock()
    type(session).id = PropertyMock(return_value=int(id_to_search))
    session.endpoint = endpoint_to_search

    manager.add_session(name_to_search, session)
    name = manager.get_session_name_by_id_endpoint(id_to_search, endpoint_to_search)
    assert_equals(name_to_search, name)


def test_get_session_id_for_client_not_there():
    manager = get_session_manager()
    manager.get_sessions_list = MagicMock(return_value=[])

    id = manager.get_session_id_for_client("name")

    assert id is None


def get_session_manager():
    ipython_display = MagicMock()
    return SessionManager(ipython_display)



# Copyright (c) 2015  aggftw@gmail.com
# Distributed under the terms of the Modified BSD License.


class LinearRetryPolicy(object):
    """Retry policy that always returns the same number of seconds to sleep between calls,
    takes all status codes 500 or above to be retriable, and retries a given maximum number of times."""

    def __init__(self, seconds_to_sleep, max_retries):
        self._seconds_to_sleep = seconds_to_sleep
        self.max_retries = max_retries

    def should_retry(self, status_code, error, retry_count):
        if None in (status_code, retry_count):
            return False
        return (status_code >= 500 and retry_count <= self.max_retries) or error

    def seconds_to_sleep(self, retry_count):
        return self._seconds_to_sleep



"""Class for implementing a Kerberos authenticator for SparkMagic"""

from requests_kerberos import HTTPKerberosAuth
import sparkmagic.utils.configuration as conf
from .customauth import Authenticator


class Kerberos(HTTPKerberosAuth, Authenticator):
    """Kerberos authenticator for SparkMagic"""

    def __init__(self, parsed_attributes=None):
        """Initializes the Authenticator with the attributes in the attributes
        parsed from a %spark magic command if applicable, or with default values
        otherwise.

        Args:
            self,
            parsed_attributes (IPython.core.magics.namespace): The namespace object that
            is created from parsing %spark magic command.
        """
        HTTPKerberosAuth.__init__(self, **conf.kerberos_auth_configuration())
        Authenticator.__init__(self, parsed_attributes)

    def __call__(self, request):
        return HTTPKerberosAuth.__call__(self, request)

    def __hash__(self):
        return hash((self.url, self.__class__.__name__))



import asyncio
import sys
import traceback
from sparkmagic.utils.constants import EXPECTED_ERROR_MSG, INTERNAL_ERROR_MSG


# == EXCEPTIONS ==
class LivyClientLibException(Exception):
    """Base class for all LivyClientLib exceptions. All exceptions that are
    explicitly raised by code in this package should be a subclass of
    LivyClientLibException. If you need to account for a new error condition,
    either use one of the existing LivyClientLibException subclasses, or create
    a new subclass with a descriptive name and add it to this file.

    We distinguish between "expected" errors, which represent errors
    that a user is likely to encounter in normal use, and "internal"
    errors, which represents exceptions that happen due to a bug in the
    library. Check EXPECTED_EXCEPTIONS to see which exceptions are
    considered "expected".
    """


class HttpClientException(LivyClientLibException):
    """An exception thrown by the HTTP client when it fails to make a
    request."""


class LivyClientTimeoutException(LivyClientLibException):
    """An exception for timeouts while interacting with Livy."""


class DataFrameParseException(LivyClientLibException):
    """An internal error which suggests a bad implementation of dataframe
    parsing from JSON -- if we get a JSON parsing error when parsing the
    results from the Livy server, this exception is thrown."""


class LivyUnexpectedStatusException(LivyClientLibException):
    """An exception that will be shown if some unexpected error happens on the
    Livy side."""


class SessionManagementException(LivyClientLibException):
    """An exception that is thrown by the Session Manager when it is a given
    session name is invalid in some way."""


class BadUserConfigurationException(LivyClientLibException):
    """An exception that is thrown when configuration provided by the user is
    invalid in some way."""


class BadUserDataException(LivyClientLibException):
    """An exception that is thrown when data provided by the user is invalid in
    some way."""


class SqlContextNotFoundException(LivyClientLibException):
    """Exception that is thrown when the SQL context is not found."""


class SparkStatementException(LivyClientLibException):
    """Exception that is thrown when an error occurs while parsing or executing
    Spark statements."""


# It has to be a KeyboardInterrupt to interrupt the notebook
class SparkStatementCancelledException(KeyboardInterrupt):
    """Exception that is thrown when a Spark statement is cancelled."""

    _orig_show_tb = None

    def __init__(self, msg):
        self.msg = msg

        # Patch _showtraceback() to avoid printing the traceback
        ipython = get_ipython()
        if SparkStatementCancelledException._orig_show_tb is None:
            SparkStatementCancelledException._orig_show_tb = ipython._showtraceback
        ipython._showtraceback = SparkStatementCancelledException._show_tb

    def __str__(self):
        return self.msg

    @staticmethod
    def _show_tb(exc_type, exc_val, tb):
        if exc_type == SparkStatementCancelledException:
            print(exc_val, file=sys.stderr)
        else:
            SparkStatementCancelledException._orig_show_tb(exc_type, exc_val, tb)


class SparkStatementCancellationFailedException(KeyboardInterrupt):
    """Exception that is thrown when a Spark statement is interrupted but fails
    to be cancelled in Livy."""


# == DECORATORS FOR EXCEPTION HANDLING ==
EXPECTED_EXCEPTIONS = [
    BadUserConfigurationException,
    BadUserDataException,
    LivyUnexpectedStatusException,
    SqlContextNotFoundException,
    HttpClientException,
    LivyClientTimeoutException,
    SessionManagementException,
    SparkStatementException,
]


def handle_expected_exceptions(f):
    """A decorator that handles expected exceptions. Self can be any object with
    an "ipython_display" attribute.
    Usage:
    @handle_expected_exceptions
    def fn(self, ...):
        etc..."""
    from sparkmagic.utils import configuration as conf

    exceptions_to_handle = tuple(EXPECTED_EXCEPTIONS)

    # Notice that we're NOT handling e.DataFrameParseException here. That's because DataFrameParseException
    # is an internal error that suggests something is wrong with LivyClientLib's implementation.
    def wrapped(self, *args, **kwargs):
        try:
            out = f(self, *args, **kwargs)
        except exceptions_to_handle as err:
            if conf.all_errors_are_fatal():
                raise err

            # Do not log! as some messages may contain private client information
            self.ipython_display.send_error(EXPECTED_ERROR_MSG.format(err))
            return None
        else:
            return out

    wrapped.__name__ = f.__name__
    wrapped.__doc__ = f.__doc__
    return wrapped


def wrap_unexpected_exceptions(f, execute_if_error=None):
    """A decorator that catches all exceptions from the function f and alerts the user about them.
    Self can be any object with a "logger" attribute and a "ipython_display" attribute.
    All exceptions are logged as "unexpected" exceptions, and a request is made to the user to file an issue
    at the Github repository. If there is an error, returns None if execute_if_error is None, or else
    returns the output of the function execute_if_error.
    Usage:
    @wrap_unexpected_exceptions
    def fn(self, ...):
        ..etc"""
    from sparkmagic.utils import configuration as conf

    def handle_exception(self, e):
        self.logger.error(
            "ENCOUNTERED AN INTERNAL ERROR: {}\n\tTraceback:\n{}".format(
                e, traceback.format_exc()
            )
        )
        self.ipython_display.send_error(INTERNAL_ERROR_MSG.format(e))
        return None if execute_if_error is None else execute_if_error()

    def wrapped(self, *args, **kwargs):
        try:
            out = f(self, *args, **kwargs)
        except Exception as err:
            if conf.all_errors_are_fatal():
                raise err
            return handle_exception(self, err)
        else:
            return out

    wrapped.__name__ = f.__name__
    wrapped.__doc__ = f.__doc__
    return wrapped


# async_wrap_unexpected_exceptions was created to handle async behavior ipykernel >=6
# It was safer to create a separate async wrapper than modify the original to accommodate both use-cases
def async_wrap_unexpected_exceptions(f, execute_if_error=None):
    """A decorator that catches all exceptions from the async function f and alerts the user about them.
    Self can be any object with a "logger" attribute and a "ipython_display" attribute.
    All exceptions are logged as "unexpected" exceptions, and a request is made to the user to file an issue
    at the Github repository. If there is an error, returns None if execute_if_error is None, or else
    returns the output of the function execute_if_error.
    Usage:
    @async_wrap_unexpected_exceptions
    async def fn(self, ...):
        ..etc"""
    from sparkmagic.utils import configuration as conf

    async def handle_exception(self, e):
        self.logger.error(
            "ENCOUNTERED AN INTERNAL ERROR: {}\n\tTraceback:\n{}".format(
                e, traceback.format_exc()
            )
        )
        self.ipython_display.send_error(INTERNAL_ERROR_MSG.format(e))
        if execute_if_error is None:
            return None

        result = execute_if_error()
        if asyncio.iscoroutine(result):
            return await result

        return result

    async def wrapped(self, *args, **kwargs):
        try:
            out = f(self, *args, **kwargs)
            if asyncio.iscoroutine(out):
                out = await out
        except Exception as err:
            if conf.all_errors_are_fatal():
                raise err
            return await handle_exception(self, err)
        else:
            return out

    wrapped.__name__ = f.__name__
    wrapped.__doc__ = f.__doc__
    return wrapped
